//	This file is part of Overpopulation, a gamescript for OpenTTD
//	by Skarfester based on City Growth Limiter Minimal by gnrd and
//	City Growth Limiter v6 by Sylf.
//
//	Copyright (C) 2023 Skarfester
//	Copyright (C) 2015 gnrd
//	Copyright (C) 2014 Sylf
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program; if not, write to the Free Software Foundation, Inc.,
//	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

require("town.nut");

class OverPop extends GSController 
{
	// Variables
	startingYear = 0;
	startingPop = 0;
	numberOfTowns = 0;
	isDataLoaded = false;
	bookPage = 1;
	goalLast = 0;
	goalNext = 1;
	goalTown = 2;
	goalTax = 3;
	badService = 0;
	growthLevel = 1;
	prevBankBal = 0;
	
	constructor() 
	{
	}
}

function OverPop::Start() 
{	
	GSLog.Info("Starting Overpopulation script...");	
	
	// Town growth rate must be non zero
	if (GSGameSettings.IsValid("town_growth_rate")) {
		if (!GSGameSettings.GetValue("town_growth_rate")) {
			GSLog.Error("Town growth is deactivated. Script terminating...");
			return;
		}
	}
	
	// Ckeck PAX cargo ID
	if (!GSCargo.IsValidCargo(0)) {
		GSLog.Error("Default PAX cargo ID is invalid. Script terminating...");
		return;
	}
	if (!GSCargo.HasCargoClass(0,  GSCargo.CC_PASSENGERS)) {
		GSLog.Error("PAX cargo ID is different from default. Script terminating...");
		return;
	}
	
	// Disable founding towns
	if (GSGameSettings.IsValid("economy.found_town")) GSGameSettings.SetValue("economy.found_town", 0);

	// Disable fund new buildings
	if (GSController.GetSetting("FundBuildings") == 0) {
		GSGameSettings.SetValue("economy.fund_buildings", 0);
	}
	
	// Create towns list
	local towns = this.CreateTownList();
	
	// Check town number
	local i = 0;
	foreach (town in towns) i++;
	numberOfTowns = i;
	GSLog.Info("Number of towns: " + numberOfTowns);
	if (i < 2) {
		GSLog.Error("Not enough towns for PAX game! Script terminating");
		return;
	}
	
	// Wait for world generation to finish
	this.Sleep(1);

	// Run these only once when new game starts	
	if (isDataLoaded == false) {	
		// Starting year (company founding year)
		startingYear = GSDate.GetYear(GSDate.GetCurrentDate());
		GSLog.Info("Company founded in " + startingYear);

		// Starting population
		startingPop = worldPop();

		// Generate story pages and goals
		storyPages();
	}
	GSLog.Info("Preparations finished. Entering main loop...");
	i = null;
	local j = null;
	local k = null;
	
	// MAIN LOOP
	while(true) {
		i = this.GetTick();
		local startMonth = GSDate.GetMonth(GSDate.GetCurrentDate());
		//GSLog.Info("Start month: " + startMonth);
		badService = 0; // reset
		foreach (town in towns) {
			if (town.ManageTown()) badService++;
			//this.Sleep(1);
		}
		GSLog.Info("Towns with insufficient service: " + badService);
		if (GSController.GetSetting("EnableTaxes")) {
			incomeTax();
		}
		checkPop();
		j = this.GetTick() - i;
		GSLog.Info("Ticks needed for processing the whole loop: " + j);
		if (j >= 2220) GSLog.Warning("The script loop took more than 30 game days so there is a risk of months being skipped");
		this.Sleep(max(1, 370 - j)); // Wait for at least 5 game days (74 * 5)
		local currentMonth = GSDate.GetMonth(GSDate.GetCurrentDate());
		//GSLog.Info("Current month: " + currentMonth);

		// wait until we move to a new month
		while(startMonth == currentMonth) {
			this.Sleep(370); // wait 5 more days
			currentMonth = GSDate.GetMonth(GSDate.GetCurrentDate());
			//GSLog.Info("Current month: " + currentMonth);
		}
 
	}
}

function OverPop::CreateTownList()
{
	local result = [];
	// Make a squirrel array of all towns.
	// For each town, an instance of Town is created
	// to store the data related to that town
	local all_towns = GSTownList();
	foreach(t, _ in all_towns)
	{
		result.append(Town(t));
	}
	return result;
}

// Apply an income tax to company id 0 (single player)
function incomeTax()
{
	local bankBal = GSCompany.GetBankBalance(0);
	local loan = getLoan();
	//GSLog.Info("Bank balance: " + bankBal + ", loan: " + loan + ", previous bank balance: " + prevBankBal);
	local tax = (growthLevel - 1) * 2;
	if (tax > 90) tax = 90; // max tax 90%
	local profit = (bankBal - loan) - prevBankBal;
	if (profit < 0) profit = 0;
	local taxes = (profit * tax) / 100;
	taxes = taxes.tointeger();
	GSLog.Info("Profit: " + profit + ", income tax: " + tax + "%, amount charged: " + taxes);
	GSCompany.ChangeBankBalance(0, -taxes, GSCompany.EXPENSES_OTHER);
	prevBankBal = GSCompany.GetBankBalance(0) - loan;
	if (prevBankBal < 0) prevBankBal = 0;

	local text = GSText(GSText.STR_GOAL_2, taxes); // update tax goal	
	GSGoal.SetText(goalTax, text);
}

// Get the current loan amount
function getLoan() {
        local mode = GSCompanyMode(0); // For some reason this needs to be in a separate function
	return GSCompany.GetLoanAmount();
}

function worldPop() { // Function taken from Deliverance by Pikka (GPL 2)

	local pop = 0;
	local townlist = GSTownList();
	
	foreach(town, z in townlist) {
		pop = pop + GSTown.GetPopulation(town);
	}
	
	GSLog.Info("Current World population: " + pop);
	
	return pop;
}

// Checks if population goal is achieved
function checkPop() {
	local popGoal = startingPop * growthLevel * 2;
	local pop = worldPop();
	local goalPer = (pop * 100) / popGoal;
	GSLog.Info("Population goal: " + popGoal + " (" + goalPer + "%)");

	// update % next goal	
	local text = GSText(GSText.STR_GOAL_0, popGoal, goalPer);
	GSGoal.SetText(goalNext, text);

	// update number of towns insufficiently serviced
	local text = GSText(GSText.STR_GOAL_3, badService);
	GSGoal.SetText(goalTown, text);	

	// check goal
	if (pop >= popGoal) {
		booknote(); // update booknote
		growthLevel ++;
	}
}

// Generate story pages and goals
function storyPages() {
	local introPage = GSStoryPage.New(0, GSText(GSText.STR_TITLE_0)); // intro page
	local text = GSText(GSText.STR_INTRO_0);
	GSStoryPage.NewElement(introPage,GSStoryPage.SPET_TEXT, 0, text);

	bookPage = GSStoryPage.New(0, GSText(GSText.STR_TITLE_1)); // notebook page
	local text = GSText(GSText.STR_BOOK_0, numberOfTowns, startingPop);
	GSStoryPage.NewElement(bookPage,GSStoryPage.SPET_TEXT, 0, text);

	local text = " "; // initiate goals
	goalLast = GSGoal.New(0,text,GSGoal.GT_NONE,0);
	local pop = startingPop * 2;	
	local text = GSText(GSText.STR_GOAL_0, pop, 50);	
	goalNext = GSGoal.New(0,text,GSGoal.GT_NONE,0);
	local text = GSText(GSText.STR_GOAL_3, badService);
	goalTown = GSGoal.New(0,text,GSGoal.GT_NONE,0);

	if (GSController.GetSetting("EnableTaxes")) { // If taxes enabled add this
		local text = GSText(GSText.STR_INTRO_1);
		GSStoryPage.NewElement(introPage,GSStoryPage.SPET_TEXT, 0, text);
		local text = GSText(GSText.STR_BOOK_1);
		GSStoryPage.NewElement(bookPage,GSStoryPage.SPET_TEXT, 0, text);
		local text = GSText(GSText.STR_GOAL_2, 0, 0);	
		goalTax = GSGoal.New(0,text,GSGoal.GT_NONE,0);	
	}
	local text = GSText(GSText.STR_INTRO_2, GSController.GetSetting("town_size_threshold"), GSController.GetSetting("min_transport_pax"));
	GSStoryPage.NewElement(introPage,GSStoryPage.SPET_TEXT, 0, text);
	GSStoryPage.Show(introPage);
}

// Update story pages
function booknote() {
	local currentYear = GSDate.GetYear(GSDate.GetCurrentDate());
	local yPassed = currentYear - startingYear;
	local grow = growthLevel * 2;
	if (GSController.GetSetting("EnableTaxes")) {
		local text = GSText(GSText.STR_BOOK_3, currentYear, grow, yPassed, grow);
		GSStoryPage.NewElement(bookPage,GSStoryPage.SPET_TEXT, 0, text);	
	} else {
		local text = GSText(GSText.STR_BOOK_2, currentYear, grow, yPassed);
		GSStoryPage.NewElement(bookPage,GSStoryPage.SPET_TEXT, 0, text);
		}

	local text = GSText(GSText.STR_GOAL_1, grow); // update last goal achieved
	GSGoal.SetText(goalLast, text);

	breakingNews(grow); // Show news
}

// Show news
function breakingNews(grow) {
	local text = null;
	if (grow == 2) {
		text = GSText(GSText.STR_NEWS_0, startingYear);
	} else {
		if (grow == 4 ) {
		text = GSText(GSText.STR_NEWS_1, startingYear);
		} else  {
			text = GSText(GSText.STR_NEWS_2, grow, startingYear);
		}
	}
	GSNews.Create(GSNews.NT_GENERAL, text, GSCompany.COMPANY_INVALID);
}

function OverPop::Save()
{
	GSLog.Info("Saving data...")

	local Save = {
	startingYear = 0,
	startingPop = 0,
	growthLevel = 0,
	prevBankBal = 0,
	}

	Save.startingYear = startingYear;
	Save.startingPop = startingPop;
	Save.growthLevel = growthLevel;
	Save.prevBankBal = prevBankBal;

	return Save;	
}

function OverPop::Load(version, data)
{
	
	GSLog.Info("Loading data...");
	
	startingYear = data.startingYear;
	startingPop = data.startingPop;
	growthLevel = data.growthLevel;
	prevBankBal = data.prevBankBal;

	isDataLoaded = true;

	GSLog.Info("Year: " + startingYear + ", Population: " + startingPop + ", Previous bank balance: " + prevBankBal);
}
